using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO.Ports;
using ZedGraph;
using System.Threading;

namespace PowerMon476
{
    public partial class MainForm : Form
    {
        public MainForm()
        {
            try
            {
                InitializeComponent();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void MainForm_Load(object sender, EventArgs e)
        {
            try
            {
                GraphPane gp = zedGraphControl.GraphPane;
                gp.Title.Text = "ECE 476 Power Monitor";
                gp.XAxis.Title.Text = "Time (s)";
                gp.YAxis.Title.Text = "Power (watts)";

                RollingPointPairList list = new RollingPointPairList(1200);
                LineItem curve = gp.AddCurve("Power", list, Color.Blue, SymbolType.Square);

                // Just manually control the X axis range so it scrolls continuously
                // instead of discrete step-sized jumps
                gp.XAxis.Scale.Min = 0;
                gp.XAxis.Scale.Max = 30;
                gp.XAxis.Scale.MinorStep = 1;
                gp.XAxis.Scale.MajorStep = 5;

                gp.YAxis.Scale.Min = 0;

                baseTime_ = (Environment.TickCount) / 1000.0;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void openConnectionToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenConnectionForm ocf = new OpenConnectionForm();
            DialogResult result = ocf.ShowDialog();

            if (result == DialogResult.OK)
            {
                try
                {
                    serialPort_ = new SerialPort(ocf.Port, 9600, Parity.None, 8, StopBits.One);
                    serialPort_.Open();
                    serialPort_.Handshake = Handshake.None;
                    serialPort_.NewLine = "\n\r";

                    GraphPane gp = zedGraphControl.GraphPane;
                    gp.XAxis.Scale.Min = 0;
                    gp.XAxis.Scale.Max = 30;
                    gp.XAxis.Scale.MinorStep = 1;
                    gp.XAxis.Scale.MajorStep = 5;

                    int unixTime = (int)GetUnixTime();
                    serialPort_.WriteLine("TIME " + unixTime);

                    baseTime_ = unixTime;
    
                    serialPort_.ReadLine();

                    serialThread_ = new Thread(new ThreadStart(SerialIOThread));
                    serialThread_.IsBackground = true;
                    serialThread_.Start();
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Unable to open device. " + ex.Message, "Device Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        private double GetUnixTime()
        {
            return (int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds;
        }

        private void AddPoint(double time, double value)
        {
            GraphPane gp = zedGraphControl.GraphPane;

            // 3 seconds per cycle
            IPointListEdit list = gp.CurveList[0].Points as IPointListEdit;

            time = time - baseTime_;

            list.Add(time, value);

            Scale xScale = gp.XAxis.Scale;
            if (time > xScale.Max - xScale.MajorStep)
            {
                xScale.Max = time + xScale.MajorStep;
                xScale.Min = xScale.Max - 30.0;
            }

            // Make sure the Y axis is rescaled to accommodate actual data
            zedGraphControl.AxisChange();
            // Force a redraw
            zedGraphControl.Invalidate();
        }

        private void SerialIOThread()
        {
            bool firstRun = true;
            while (true)
            {
                try
                {
                    string line = serialPort_.ReadLine();

                    if (line == null) continue;
                    //Random r = new Random();
                    //string line = (Environment.TickCount) / 1000.0 + ":V" + (2.5*r.NextDouble() + 40) + ",F" + (0.5*r.NextDouble() + 60);
                    //string line = (Environment.TickCount) / 1000.0 + ":V" + 270 * Math.Sin(2.0 * Math.PI * (Environment.TickCount) / 1000.0 / 3.0);
                    double time;
                    double.TryParse(line.Substring(0, line.IndexOf(":")), out time);
                    //MessageBox.Show("Time: " + time + ", Base: " + baseTime_);
                    if (firstRun)
                    {
                        firstRun = false;
                        baseTime_ = time;
                    }

                    string[] lineArray = line.Substring(line.IndexOf(":") + 1).Split(new char[] { ',' });

                    foreach (string data in lineArray)
                    {
                        char type = data[0];
                        double value;
                        double.TryParse(data.Substring(1), out value);
                        switch (type)
                        {
                            case 'P':
                                AddPoint(time, value);
                                ChangeLabel(powerLabel, "Power: " + value.ToString("  0.00") + " W");
                                break;
                            case 'F':
                                ChangeLabel(freqLabel, "Frequency: " + value.ToString("  0.00") + " Hz");
                                break;
                            case 'I':
                                ChangeLabel(currentLabel, "Current: " + value.ToString("  0.00") + " A");
                                break;
                            case 'V':
                                ChangeLabel(voltageLabel, "Voltage: " + value.ToString("  0.00") + " V");
                                break;
                            case 'E':
                                ChangeLabel(energyLabel, "Energy: " + value.ToString("   0.0000") + " KW*Hrs");
                                break;
                            case 'A':
                                ChangeLabel(appPowerLabel, "App. Power: " + value.ToString("  0.00") + " VA");
                                break;
                            case 'R':
                                ChangeLabel(powerFactorLabel, "Power Factor: " + value.ToString("  0.00"));
                                break;
                        }
                    }
                }
                catch (Exception ex)
                {
                    //MessageBox.Show(ex.Message);
                }
            }
        }

        private void ChangeLabel(System.Windows.Forms.Label label, string text)
        {
            if (InvokeRequired)
            {
                Invoke(new MethodInvoker(delegate() { label.Text = text; }));
                return;
            }

            label.Text = text;
        }

        private void closeConnectionToolStripMenuItem_Click(object sender, EventArgs e)
        {
            try
            {
                serialPort_.Close();
                serialThread_.Abort();
            }
            catch (Exception)
            {
            }

            serialPort_ = null;
        }

        private Thread serialThread_;
        private double baseTime_;
        private SerialPort serialPort_;

        private void relayOnToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (serialPort_ == null)
            {
                MessageBox.Show("Device not connected.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            serialPort_.WriteLine("RELAY 1");
        }

        private void relayOffToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (serialPort_ == null)
            {
                MessageBox.Show("Device not connected.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            serialPort_.WriteLine("RELAY 0");
        }

        private void setThresholdToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (serialPort_ == null)
            {
                MessageBox.Show("Device not connected.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            SetLimitForm slf = new SetLimitForm();
            DialogResult result = slf.ShowDialog();
            if (result == DialogResult.OK)
            {
                serialPort_.WriteLine("THRESHOLD " + slf.Limit);
            }
        }
    }
}