/*
 *  final.c
 *
 *
 *  Created by Cliff Jao and Xi Guo on 4/14/08.
 *
 * Voltage on A0
 * Current on A1
 *
 * Serial commands:
 * SET_TIME <long seconds>
 * RELAY <0/1>
 * SET_THRESHOLD <watts>
 */

#include <Mega32.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#define SAMPLES 1000
#define t1 1 // 1 ms.
#define t2 1000 // 500 ms
#define t3 1 // 1 ms
#define R_SENSE 0.2

#define Vref 5.0

char *cmds[] = {"TIME", "RELAY", "THRESHOLD"};
#define TIME 0
#define RELAY 1
#define THRESHOLD 2

long FP_ADC_CONV, FP_OFFSET; 

////// 20:12 fixed point macros
#define int2fix(a)   (((long)(a))<<12)    //Convert char to fix. a is a char
#define float2fix(a) ((long)((a)*4096.0)) //Convert float to fix. a is a float
#define fix2float(a) ((float)(a)/4096.0)  //Convert fix to float. a is an int
#define multfix(a,b) ((long)((((long)(a))*((long)(b)))>>12))

long fpVSquareSum;
long fpISquareSum;
long fpPowerSum;
float rPower, aPower;
float powerFactor;

float iScaleFactor, vScaleFactor, rpScaleFactor;

///// Periods //////
long fpVLastValue;
unsigned long vPeriodSum;
int vLastZeroMsec;
unsigned int vPeriodCount;


///// Watt Hour Stuff ////
float kWattHrs;

unsigned char r_index, r_buffer[40], r_ready, r_char;
unsigned char t_index, t_buffer[350], t_ready, t_char;
void gets_init();
void puts_init();

char time0, time1, time2, time3;
unsigned int sample;

// seconds since January 1, 1970 (UNIX time)
unsigned long unixtime;
unsigned int msecs;
unsigned long offTime;
unsigned long onTime;

long threshold;
char shutdown;

void task1();
void task2();
void task3();
void initialize();
int fpMult(int a, int b);


void main() {
  initialize();

  while(1) {
    if (time1 == 0) task1();
    if (time2 == 0) task2();
    if (time3 == 0) task3();
  }
}

void task3() { 

  int int_param, i;
  char cmd;
  char cmdStr[50];
  char param[50];
  char *space;
  int spaceIndex;

  time3 = t3;
  
  if (r_ready != 1) return;
  
  // try to find space
  space = strchr(r_buffer, ' ');
  
  if (space != NULL) {
    // space was found, extract param
    spaceIndex = space - r_buffer;
    strncpy(cmdStr, r_buffer, spaceIndex);
    cmdStr[spaceIndex] = '\0';
    strcpy(param, space + 1);     
  }
  else
  {
    // space not found, just copy over
    strcpy(cmdStr, r_buffer);
  }
  
      // cmd will be -1 if no command
    cmd = -1;
    for (i = 0; i < 3; i++) {
      if (!strcmp(cmdStr, cmds[i]))
      {
        cmd = i;
        break;
      }
    }
    
    switch(cmd) {
      case TIME:
        sscanf(param, "%ld", &unixtime);
        msecs = 0;
        break;
      case RELAY:
        sscanf(param, "%d", &int_param);
        if (int_param) shutdown = 0;
        else shutdown = 1;
        break;
      case THRESHOLD:
        sscanf(param, "%ld", &threshold);
        break;
    }
    
    //printf("CMD OK\n\r");
    gets_init();
}

void task2() {
  time2 = t2;
  PORTD.2 = ~PORTD.2;
}

void task1() {
  char Ain0, Ain1;
  long fpVin0, fpVin1, fpVin, fpIin;
  long fpAin0, fpAin1;
  float vAvg, iAvg;
  int vPeriod;

  time1 = t1;

  // start conversion for channel 0
  ADMUX.0 = 0;
  ADCSR.6 = 1;

  // wait for conversion to finish
  while (ADCSR.6);

  // read Ain0;
  Ain0 = ADCH;

  // start conversion for channel 1
  ADMUX.0 = 1;
  ADCSR.6 = 1;

  // wait for conversion to finish
  while (ADCSR.6);

  Ain1 = ADCH;

  // go to 20:12 fixed point
  fpAin0 = (long)Ain0 << 12;
  fpAin1 = (long)Ain1 << 12;

  fpVin0 = multfix(fpAin0, FP_ADC_CONV) - FP_OFFSET;
  fpVin1 = multfix(fpAin1, FP_ADC_CONV) - FP_OFFSET;

  //fpVin = multfix(fpVin0, FP_V_CONV); // convert to actual voltage
  //fpVin1 = multfix(fpVin1, FP_I_CONV); // convert to actual current

  fpPowerSum += multfix(fpVin0, fpVin1);
  fpVSquareSum += multfix(fpVin0, fpVin0);
  fpISquareSum += multfix(fpVin1, fpVin1);

  ///////// FREQUENCY CALCULATION //////
  if (fpVLastValue < 0 && fpVin0 >= 0) {
    // found a zero crossing from neg. to pos.!
    vPeriod = msecs - vLastZeroMsec;

    if (vPeriod > 0) {
      vPeriodSum += vPeriod;
      vPeriodCount++;
    }

    //ch1PeriodCount = 0;
    vLastZeroMsec = msecs;

  }

  fpVLastValue = fpVin0;
  ///////////////////////////////////


  sample++;

  if (sample == SAMPLES) {
    // this only happens once a second,
    // so floating point mults should be okay
    vAvg = (float) sqrt(vScaleFactor * fpVSquareSum);
    iAvg = (float) sqrt(iScaleFactor * fpISquareSum);
    rPower = (float) rpScaleFactor * fpPowerSum; // watts
    aPower = vAvg * iAvg; // VA
    powerFactor = rPower / aPower;
    kWattHrs += rPower / 3600000.0;

    //if (power < 10) power = 0; // probably just noise

    sprintf(t_buffer, "%ld:P%3.2f,A%3.2f,I%3.2f,V%3.2f,F%3.2f,E%f,R%3.2f\n\r", unixtime, rPower, aPower, iAvg, vAvg, 1000.0 * (float)vPeriodCount / (float)vPeriodSum, kWattHrs, powerFactor);
    puts_init();
    
    if (rPower > threshold) shutdown = 1;
    
    if (shutdown) PORTD.3 = 0; // turn relay off
    else PORTD.3 = 1;

    // reset for next sample
    fpVSquareSum = 0;
    fpISquareSum = 0;
    fpPowerSum = 0;
    vPeriodSum = 0;
    vPeriodCount = 0;
    sample = 0;
  }
}

interrupt [TIM0_COMP] void timer0_compare(void) {
  if (time1 > 0) --time1;
  if (time2 > 0) --time2;
  if (time3 > 0) --time3;

  msecs++;
  if (msecs == 1000)
  {
    msecs = 0;
    unixtime++;
  }
}

void initialize() {
  long t;

  /////////////// TIMER 0 ///////////////////////
  TIMSK = 2; // turn on timer 0 compare match ISR
  OCR0 = 250; // sets compare to 250 ticks
  TCCR0 = 0b00001011; // sets prescalar to 64

  //set up UART
  UCSRB = 0x18;
  UBRRL = 103;

  //printf("Starting...\n\rEnter UNIX time: ");

  //scanf("%ld", &t);
  //printf("\n\rYou entered: %ld\r\n", t);

  unixtime = 0;
  msecs = 0;

  DDRD = 0xFF; // output
  PORTD.2 = 0;
  PORTD.3 = 1; // turn relay on

  threshold = 1000;
  shutdown = 0;

  kWattHrs = 0;
  time1 = t1;
  time2 = t2;

  vScaleFactor = 1002001.0 / (4096.0 * SAMPLES);
  iScaleFactor = 1 / (R_SENSE * R_SENSE) / (4096.0 * SAMPLES);
  rpScaleFactor = 1001.0 / R_SENSE / (4096.0 * SAMPLES);

  // internal Aref = Vcc, left adjust result
  ADMUX = 0b01100000;
  //enable ADC and set prescaler to 1/128*16MHz=125,000
  //and clear interupt enable
  ADCSR = 0b11000111;

  // calculate conversion factor
  //FP_ADC_CONV = float2fix(Vref / 256.0);
  FP_ADC_CONV = float2fix(0.512 / 256.0); // opto
  //FP_I_CONV = float2fix(1 / R_SENSE);

  FP_OFFSET = float2fix(0.256);
  #asm("sei")
  
  gets_init();
}

//**********************************************************
//UART character-ready ISR
interrupt [USART_RXC] void uart_rec() {
	r_char=UDR;    //get a char
	//UDR=r_char;    //then print it
	//build the input string
	if (r_char != '\r') r_buffer[r_index++]=r_char;
	else {
		putchar('\n');  			//use putchar to avoid overwrite
		r_buffer[r_index]=0x00;    	//zero terminate
		r_ready=1;				   	//signal cmd processor
		UCSRB.7=0;   				//stop rec ISR
	}
}

/**********************************************************/
//UART xmit-empty ISR
interrupt [USART_DRE] void uart_send() {
	t_char = t_buffer[++t_index];
	if (t_char == 0) {
		UCSRB.5=0; //kill isr
		t_ready=1; //transmit done
	}
	else UDR = t_char ;     //send the char
}

//**********************************************************
//  -- non-blocking keyboard check initializes ISR-driven
// receive. This routine merely sets up the ISR, which then
//does all the work of getting a command.
void gets_init() {
  r_ready=0;
  r_index=0;
  UCSRB.7=1;
}

//**********************************************************
//  -- nonblocking print: initializes ISR-driven
// transmit. This routine merely sets up the ISR, then
//send one character, The ISR does all the work.
void puts_init()  {
  t_ready=0;
  t_index=0;
  if (t_buffer[0]>0) {
  	putchar(t_buffer[0]);
  	UCSRB.5=1;
  }
}
